package org.apidesign.apifest08.currency;

import static org.apidesign.apifest08.currency.Assert.notNull;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.Currency;
import java.util.List;

public final class Convertor {
	
	private List<ConvertorDelegate> convertorDelegates = new ArrayList<ConvertorDelegate>();
	
	
	/**
	 * Create new instance of the converter for the given currencies and its rate.
	 * 
	 * @param rateValue the rate between the first and the second currency
	 * @param currencyFirst the first currency
	 * @param currencySecond the second currency
	 */
	public Convertor(BigDecimal rateValue, Currency currencyFirst, Currency currencySecond) {
		notNull(currencyFirst, "currencyFirst");
		notNull(currencySecond, "currencySecond");		
		notNull(rateValue, "rateValue");	
		convertorDelegates.add(new ConvertorDelegate(rateValue, currencyFirst, currencySecond));
	}
	
	/**
	 * Create new instance of the convertor from the given convertors. 
	 * @param convertors the convertors
	 */
	public Convertor(Convertor... convertors) {
		notNull(convertors, "convertors");
		if(convertors.length == 0) {
			throw new IllegalArgumentException("There must be at least one converter.");
		}
		
		for(Convertor convertor: convertors) {
			if(convertor != null) {
				convertorDelegates.addAll(convertor.convertorDelegates);
			}
		}
	}
    	
	/**
	 * Converts an amount value between the two currencies of this converter.
	 *  
	 * @param amount an amount
	 * @param fromCurrency an amount currency
	 * @param toCurrency to a target currency
	 * @return a converted amount value
	 * 
	 * @throws ConversionException if the conversion fails
	 * @throws UnsupportedConversionException if the conversion between a given currencies is not supported.
	 */
	public Amount convert(BigDecimal amount, Currency fromCurrency, Currency toCurrency) throws ConversionException {
		notNull(amount, "amount");
		notNull(fromCurrency, "fromCurrency");
		notNull(toCurrency, "toCurrency");
		ConvertorDelegate appropriateDelegate = null;
		
		//try find an appropriate delegate for conversion
		for(ConvertorDelegate delegate : convertorDelegates) {
			if(delegate.isConversionSupported(fromCurrency, toCurrency)) {
				appropriateDelegate = delegate;
				break;
			}
		}
		
		if(appropriateDelegate == null) {
			throw new UnsupportedConversionException(fromCurrency, toCurrency);
		}
		
		return appropriateDelegate.convert(amount, fromCurrency, toCurrency);	
	}
	
	/**
	 * Internal delegate implements a logic for conversion between two currencies
	 * and vice versa.
	 * @see #isConversionSupported(Currency, Currency)
	 */
	private static class ConvertorDelegate {
		private final Currency first;
		private final Currency second;
		private final BigDecimal rateValue; // a rate between the first currency and the second currency
		public static final BigDecimal one = new BigDecimal(1);
		
		private ConvertorDelegate(BigDecimal rateValue, Currency currencyFirst, Currency currencySecond) {
			this.rateValue = rateValue;
			this.first = currencyFirst;
			this.second = currencySecond;
		}
		
		private Amount convert(BigDecimal amount, Currency fromCurrency, Currency toCurrency) throws ConversionException {
			BigDecimal rateValue = getRateValue(fromCurrency, toCurrency);
			BigDecimal result = rateValue.multiply(amount);			
			return new Amount(result, toCurrency);	
		}
		
		private BigDecimal getRateValue(Currency fromCurrency, Currency toCurrency) {		
			
			BigDecimal retVal;
			
			if(first == fromCurrency) {
				retVal = rateValue;
			} else {	
				//reverse rate	
				retVal = one.divide(rateValue, 10 ,RoundingMode.HALF_UP);
			}
			
			return retVal;
		}
		
		/**
		 * @return <code>true</code> if the delegate is able to convert from the given currency
		 * to the given currency and vice versa otherwise <code>false</code>.
		 */
		private boolean isConversionSupported(Currency fromCurrency, Currency toCurrency) {
			return ((fromCurrency == first || fromCurrency == second) && (toCurrency == first || toCurrency == second));
		}
	}
}
